from typing import Type

from scapy.all import Packet
from scapy.layers.inet import TCP, UDP

from heifip.layers.ip import IPPacket
from heifip.plugins.header import CustomTCP, CustomUDP


class TransportPacket(IPPacket):
    def __init__(self, packet: Packet, address_mapping={}, layer_map={}):
        IPPacket.__init__(self, packet, address_mapping, layer_map)

    def header_preprocessing(self):
        if TCP in self.layer_map:
            layer_copy = self.packet[TCP]
            layer_copy = self.header_preprocessing_tcp(layer_copy)
            if self.packet[TCP].payload != None:
                layer_copy.payload = self.packet[TCP].payload
            self.packet[TCP] = layer_copy

        if UDP in self.layer_map:
            layer_copy = self.packet[UDP]
            layer_copy = self.header_preprocessing_udp(layer_copy)
            if self.packet[UDP].payload != None:
                layer_copy.payload = self.packet[UDP].payload
            self.packet[UDP] = layer_copy

        super().header_preprocessing()

    def header_preprocessing_tcp(self, layer_copy: Packet):
        return CustomTCP(flags=layer_copy.flags, options=layer_copy.options)

    def header_preprocessing_udp(self, layer_copy: Packet):
        return CustomUDP()



from abc import ABC, abstractmethod

__author__ = "Stefan Machmeier"
__copyright__ = "Copyright 2023, heiFIP"
__credits__ = ["Manuel Trageser"]
__license__ = "EUPL"
__version__ = "1.1.0"
__maintainer__ = "Stefan Machmeier"
__email__ = "stefan.machmeier@uni-heidelberg.de"
__status__ = "Production"


class NetworkTrafficImage(ABC):
    def __init__(self, fill=0, dim=8) -> None:
        self.fill = fill
        self.dim = dim



import binascii
import logging

import numpy as np
from scapy.all import Packet, raw

from heifip.images import NetworkTrafficImage


class FlowImage(NetworkTrafficImage):
    def __init__(
        self,
        packets,
        dim=16,
        fill=0,
        append=False,
    ) -> None:
        NetworkTrafficImage.__init__(self, fill, dim)
        self.packets = packets
        self.append = append
        self.matrix, self.binaries = self.__get_matrix(self.dim, self.append, self.fill, self.packets)
        del packets
        

    def __get_matrix(self, dim: int, append: bool, fill: int, packets: [Packet]):
        """
            Creates a matrix of a list of Scapy Packet.
        """
        binaries = []
        for packet in self.packets:
            # get Hex data
            hexst = binascii.hexlify(raw(packet.packet))
            # Append octet as integer
            binaries.append(
                [int(hexst[i : i + 2], 16) for i in range(0, len(hexst), 2)]
            )
        fh = None
        # Append packets after another or write each packet in a row
        if append:
            fh = np.concatenate([np.array(xi) for xi in binaries])
            rn = len(fh) // dim + (len(fh) % dim > 0)
            fh = np.pad(fh, (0, (rn * dim) - fh.shape[0]), 'constant')
            fh = fh.reshape(rn, dim)
        else:
            length = max(map(len, binaries))
            fh = np.array([xi + [fill] * (length - len(xi)) for xi in binaries])

        fh = np.uint8(fh)

        return fh, binaries



import binascii
import logging

import numpy as np
from scapy.all import Packet, raw

from heifip.images import NetworkTrafficImage


class PacketImage(NetworkTrafficImage):
    def __init__(
        self,
        packet: Packet,
        dim=8,
        fill=0,
        auto_dim=False
    ) -> None:
        NetworkTrafficImage.__init__(self, fill, dim)
        self.auto_dim = auto_dim
        self.matrix, self.binaries = self.__get_matrix(self.dim, self.auto_dim, self.fill, packet)

        del packet

    def __get_matrix(self, dim: int, auto_dim: int, fill: int, packet: Packet):
        # get Hex data
        hexst = binascii.hexlify(raw(packet.packet))
        # Append octet as integer
        binaries = [int(hexst[i: i + 2], 16) for i in range(0, len(hexst), 2)]
        # Get min dim
        length = len(binaries)
        if auto_dim:
            dim = int(np.ceil(np.sqrt(length)))

        # Create array and shape it to dim
        fh = np.array(binaries + [fill] * (dim * dim - len(binaries)))
        fh = fh[0:dim * dim].reshape(dim, dim)

        fh = np.uint8(fh)

        return fh, binaries



import logging

import numpy as np
from scapy.all import Packet, chexdump, hexdump, raw

from heifip.images import NetworkTrafficImage


class MarkovTransitionMatrix(NetworkTrafficImage):
    def __init__(
        self,
    ) -> None:
        NetworkTrafficImage.__init__(self)

    def bit_array(self, packet):
        bytes_as_bits =  ''.join(format(byte, '08b') for byte in bytes(packet.packet))
        transition = []
        for i in range(0, len(bytes_as_bits), 4):
            transition.append(int(bytes_as_bits[i:i+4], 2))
        return transition

    def transition_matrix(self, transitions):
        n = 16

        M = [[0]*n for _ in range(n)]

        for (i,j) in zip(transitions,transitions[1:]):
            M[i][j] += 1

        #now convert to probabilities:
        for row in M:
            s = sum(row)
            if s > 0:
                row[:] = [f/s for f in row]
        return M

class MarkovTransitionMatrixFlow(MarkovTransitionMatrix):
    def __init__(
        self,
        packets: [Packet],
        cols: int = 4
    ) -> None:
        MarkovTransitionMatrix.__init__(self)

        result = []
        for packet in packets:
            transition = self.bit_array(packet)
            m = self.transition_matrix(transition)
            result.append(np.array(m))
        
        # Create tiled image
        fh = self.__tile_images(result, cols, 16)

        # Convert to int
        self.matrix = fh
        del packets

    def __tile_images(self, images, cols, dim):
        """Tile images of same size to grid with given number of columns.

        Args:
            images (collection of ndarrays)
            cols (int): number of colums

        Returns:
            ndarray: stitched image
        """
        k = 0
        rows = []
        for i in range(0, cols):
            row = None
            for j in range(0, cols):
                if len(images) > k:
                    im = images[k]
                else:
                    im = np.zeros((dim, dim))
                
                if row is None:
                    row = im
                else:
                    row = np.concatenate((row, im), axis=1)
                k += 1
            rows.append(row)
        tiled = np.concatenate(rows)

        return tiled

class MarkovTransitionMatrixPacket(MarkovTransitionMatrix):
    def __init__(
        self,
        packet: Packet,
    ) -> None:
        MarkovTransitionMatrix.__init__(self)

        transition = self.bit_array(packet)
        m = self.transition_matrix(transition)
        self.matrix = np.array(m)

        del packet



from typing import Type

from scapy.all import Packet, RandIP, RandIP6
from scapy.layers.inet import IP
from scapy.layers.inet6 import IPv6

from heifip.layers.packet import EtherPacket
from heifip.plugins.header import CustomIP, CustomIPv6


class IPPacket(EtherPacket):
    def __init__(self, packet: Packet, address_mapping={}, layer_map={}):
        EtherPacket.__init__(self, packet, address_mapping, layer_map)
        if IP in self.layer_map:
            self.__filter_ipv4()
        elif IPv6 in self.layer_map:
            self.__filter_ipv6()

    def __filter_ipv4(self):
        previous_src = self.packet[IP].src
        previous_dst = self.packet[IP].dst

        if previous_src in self.address_mapping:
            new_src = self.address_mapping[previous_src]
        else:
            new_src = RandIP()._fix()
            self.address_mapping[previous_src] = new_src

        if previous_dst in self.address_mapping:
            new_dst = self.address_mapping[previous_dst]
        else:
            new_dst = RandIP()._fix()
            self.address_mapping[previous_dst] = new_dst

        self.packet[IP].src = new_src
        self.packet[IP].dst = new_dst

    def header_preprocessing(self):        
        if IP in self.layer_map:
            layer_copy = self.packet[IP]
            layer_copy = self.header_preprocessing_ipv4(layer_copy)
            if self.packet[IP].payload != None:
                layer_copy.payload = self.packet[IP].payload
            self.packet[IP] = layer_copy
        if IPv6 in self.layer_map:
            layer_copy = self.packet[IPv6]
            layer_copy = self.header_preprocessing_ipv6(layer_copy)
            if self.packet[IPv6].payload != None:
                layer_copy.payload = self.packet[IPv6].payload
            self.packet[IPv6] = layer_copy

        super().header_preprocessing()

    def header_preprocessing_ipv4(self, layer_copy: Packet):
        return CustomIP(
            version=layer_copy.version,
            tos=layer_copy.tos,
            ttl=layer_copy.ttl,
            flags=layer_copy.flags,
            proto=layer_copy.proto,
        )

    def __filter_ipv6(self):
        previous_src = self.packet[IPv6].src
        previous_dst = self.packet[IPv6].dst

        if previous_src in self.address_mapping:
            new_src = self.address_mapping[previous_src]
        else:
            new_src = RandIP6()._fix()
            self.address_mapping[previous_src] = new_src

        if previous_dst in self.address_mapping:
            new_dst = self.address_mapping[previous_dst]
        else:
            new_dst = RandIP6()._fix()
            self.address_mapping[previous_dst] = new_dst

        self.packet[IPv6].src = new_src
        self.packet[IPv6].dst = new_dst

    def header_preprocessing_ipv6(self, layer_copy: Packet):
        return CustomIPv6(
            version=layer_copy.version,
            tc=layer_copy.tc,
            nh=layer_copy.nh,
            hlim=layer_copy.hlim,
        )



from typing import Type

from scapy.all import Packet
from scapy.layers.http import HTTPRequest, HTTPResponse

from heifip.layers.transport import TransportPacket
from heifip.plugins.header import (CustomHTTP, CustomHTTP_Request,
                                   CustomHTTP_Response)


class HTTPPacket(TransportPacket):
    def __init__(self, packet: Packet, address_mapping={}, layer_map={}):
        TransportPacket.__init__(self, packet, address_mapping, layer_map)
    def header_preprocessing(self):
        super().header_preprocessing()


class HTTPRequestPacket(HTTPPacket):
    def __init__(self, packet: Packet, address_mapping={}, layer_map={}):
        HTTPPacket.__init__(self, packet, address_mapping, layer_map)

    def header_preprocessing(self):
        layer_copy = self.packet[HTTPRequest]
        layer_copy = CustomHTTP_Request(
            Method=layer_copy.Method,
            Path=layer_copy.Path,
            User_Agent=layer_copy.User_Agent,
            Content_Type=layer_copy.Content_Type,
            Connection=layer_copy.Connection,
            Accept=layer_copy.Accept,
            Accept_Charset=layer_copy.Accept_Charset,
            Cookie=layer_copy.Cookie,
            TE=layer_copy.TE,
        )

        if not self.packet[HTTPRequest].payload is None:
            layer_copy.payload = self.packet[HTTPRequest].payload

        self.packet[HTTPRequest] = layer_copy

        super().header_preprocessing()


class HTTPResponsePacket(HTTPPacket):
    def __init__(self, packet: Packet, address_mapping={}, layer_map={}):
        HTTPPacket.__init__(self, packet, address_mapping, layer_map)

    def header_preprocessing(self):
        layer_copy = self.packet[HTTPResponse]
        layer_copy = CustomHTTP_Response(
            Status_Code=layer_copy.Status_Code,
            Server=layer_copy.Server,
            Content_Type=layer_copy.Content_Type,
            Connection=layer_copy.Connection,
            Content_Encoding=layer_copy.Content_Encoding,
            Set_Cookie=layer_copy.Set_Cookie,
            Transfer_Encoding=layer_copy.Transfer_Encoding,
        )

        if self.packet[HTTPResponse].payload != None:
            layer_copy.payload = self.packet[HTTPResponse].payload

        self.packet[HTTPResponse] = layer_copy

        super().header_preprocessing()




import hashlib
import os

from scapy.all import *
from scapy.all import TCP


class StrCustomTerminatorField(StrField):
    __slots__ = ["remain", "terminator", "consume_terminator"]

    def __init__(self, name, default, fmt="H", remain=0, terminator="\x00\x00", consume_terminator=True):
        super().__init__(name, default, fmt, remain)

        self.terminator = terminator
        self.consume_terminator = consume_terminator

    def addfield(self, pkt, s, val):
        return s+self.i2m(pkt, val)+self.terminator

    def getfield(self, pkt, s):
        l = s.find(self.terminator)
        if l < 0:
            # XXX terminator not found
            return "", s
        if self.consume_terminator:
            return s[l+len(self.terminator):], self.m2i(pkt, s[:l])
        return s[l:], self.m2i(pkt, s[:l])

    def randval(self):
        return RandTermString(RandNum(0, 1200), self.terminator)


class HintField(StrField):
    def __init__(self, name, default, fmt="H", remain=0):
        super().__init__(name, default, fmt, remain)

    def i2len(self, pkt, i):
        return 0

    def i2m(self, pkt, x):
        return ''


class DynamicStrField(Field):
    __slots__ = [
        "name",
        "fmt",
        "default",
        "sz",
        "owners",
        "struct",
        "remain",
        "adjust"
    ]

    def __init__(self, name, default, fmt="H", remain=0, adjust=lambda pkt, x: x):

        super().__init__(name, default, fmt)

        self.remain = remain

        self.adjust = adjust

    def i2len(self, pkt, i):
        return len(i)

    def i2m(self, pkt, x):
        if x is None:
            x = ""
        elif type(x) is not str:
            x = str(x)

        x = self.adjust(pkt, x)
        return x

    def addfield(self, pkt, s, val):
        return s+self.i2m(pkt, val)

    def getfield(self, pkt, s):
        if self.remain == 0:
            return "", self.m2i(pkt, s)
        else:
            return s[-self.remain:], self.m2i(pkt, s[:-self.remain])

    def randval(self):
        return RandBin(RandNum(0, 1200))


class BLenField(LenField):
    __slots__ = ["adjust", "numbytes", "length_of", "count_of"]

    def __init__(self, name, default, fmt="I", adjust=lambda pkt, x: x, numbytes=None, length_of=None, count_of=None):
        self.name = name
        self.adjust = adjust
        self.numbytes = numbytes
        self.length_of = length_of
        self.count_of = count_of
        super().__init__(name, default, fmt)

        if fmt[0] in "@=<>!":
            self.fmt = fmt
        else:
            self.fmt = "!"+fmt
        self.default = self.any2i(None, default)
        self.sz = struct.calcsize(self.fmt) if not numbytes else numbytes
        self.owners = []

    def addfield(self, pkt, s, val):
        """Add an internal value  to a string"""
        pack = struct.pack(self.fmt, self.i2m(pkt, val))
        if self.numbytes:
            pack = pack[len(pack)-self.numbytes:]
        return s+pack

    def getfield(self, pkt, s):
        """Extract an internal value from a string"""
        upack_data = s[:self.sz]
        # prepend struct.calcsize()-len(data) bytes to satisfy struct.unpack
        upack_data = '\x00'*(struct.calcsize(self.fmt)-self.sz) + upack_data

        return s[self.sz:], self.m2i(pkt, struct.unpack(self.fmt, upack_data)[0])

    def i2m(self, pkt, x):
        if x is None:
            if not (self.length_of or self.count_of):
                x = len(pkt.payload)
                x = self.adjust(pkt, x)
                return x

            if self.length_of is not None:
                fld, fval = pkt.getfield_and_val(self.length_of)
                f = fld.i2len(pkt, fval)
            else:
                fld, fval = pkt.getfield_and_val(self.count_of)
                f = fld.i2count(pkt, fval)
            x = self.adjust(pkt, f)
        return x


class XBLenField(BLenField):
    def i2repr(self, pkt, x):
        return lhex(self.i2h(pkt, x))


class XLenField(LenField):
    def i2repr(self, pkt, x):
        return lhex(self.i2h(pkt, x))


class XFieldLenField(FieldLenField):
    def i2repr(self, pkt, x):
        return lhex(self.i2h(pkt, x))


SSH_MESSAGE_TYPES = {0x01: "disconnect",
                     0x14: "kex_init",
                     0x15: "new_keys",
                     0xff: "unknown"}
SSH_TYPE_BOOL = {0x00: True,
                 0xff: False}

SSH_ALGO_CIPHERS = "none,aes128-ctr,aes192-ctr,aes256-ctr,arcfour256,arcfour128,aes128-gcm@openssh.com,aes256-gcm@openssh.com,aes128-cbc,3des-cbc,blowfish-cbc,cast128-cbc,aes192-cbc,aes256-cbc,arcfour".split(
    ",")
SSH_ALGO_HMACS = "none,hmac-md5-etm@openssh.com,hmac-sha1-etm@openssh.com,umac-64-etm@openssh.com,umac-128-etm@openssh.com,hmac-sha2-256-etm@openssh.com,hmac-sha2-512-etm@openssh.com,hmac-ripemd160-etm@openssh.com,hmac-sha1-96-etm@openssh.com,hmac-md5-96-etm@openssh.com,hmac-md5,hmac-sha1,umac-64@openssh.com,umac-128@openssh.com,hmac-sha2-256,hmac-sha2-512,hmac-ripemd160,hmac-sha1-96,hmac-md5-96".split(
    ",")
SSH_ALGO_KEX = "none,ecdh-sha2-nistp256,ecdh-sha2-nistp384,ecdh-sha2-nistp521,diffie-hellman-group-exchange-sha256,diffie-hellman-group-exchange-sha1,diffie-hellman-group14-sha1,diffie-hellman-group1-sha1".split(
    ",")
SSH_ALGO_COMPRESSION = "none,zlib,zlib@openssh.com,none".split(",")
SSH_ALGO_HOSTKEY = "none,ecdsa-sha2-nistp521,ssh-rsa,ssh-dss".split(",")


def ssh_name_list(name, fmt="!I", numbytes=None, default=''):
    return [XBLenField("%s_length" % name, None, length_of="%s" % name, fmt=fmt, numbytes=numbytes),
            StrLenField("%s" % name, default, length_from=lambda x:getattr(x, "%s_length" % name)), ]


class SSHIdent(Packet):
    name = "SSH Ident"
    fields_desc = [
        StrField("ident", "SSH-2.0-ScapySSHLayer\r\n"),
    ]


def ssh_calculate_mac(pkt, x):
    if len(x):
        return x
    if not pkt.mac in ('md5', 'sha-1'):
        return x
    return getattr(hashlib, pkt.mac)(pkt.data).digest()


class SSHEncryptedPacket(Packet):
    name = "SSH Encrypted Packet"
    fields_desc = [
        StrField("data", None),
        DynamicStrField("mac", None, adjust=ssh_calculate_mac),
        HintField("encryption", None),
        # HintField("mac",'md5'),
        HintField("compression", None),
    ]


class SSHMessage(Packet):
    name = "SSH Message"
    fields_desc = [
        XBLenField("length", None, fmt="!I", adjust=lambda pkt,
                   x: x+2 if pkt.lastlayer().haslayer(Raw) else x+2),
        XBLenField("padding_length", None, fmt="!B", adjust=lambda pkt, x: len(
            pkt.lastlayer()) if pkt.lastlayer().haslayer(Raw) else 0),
        ByteEnumField("type", 0xff, SSH_MESSAGE_TYPES),
    ]


class SSHKexInit(Packet):
    name = "SSH Key Exchange Init"
    fields_desc = [StrFixedLenField("cookie", os.urandom(16), 16), ] \
        + ssh_name_list("kex_algorithms", default=",".join(SSH_ALGO_KEX)) \
        + ssh_name_list("server_host_key_algorithms", default=",".join(SSH_ALGO_HOSTKEY)) \
        + ssh_name_list("encryption_algorithms_client_to_server", default=",".join(SSH_ALGO_CIPHERS)) \
        + ssh_name_list("encryption_algorithms_server_to_client", default=",".join(SSH_ALGO_CIPHERS)) \
        + ssh_name_list("mac_algorithms_client_to_server", default=",".join(SSH_ALGO_HMACS)) \
        + ssh_name_list("mac_algorithms_server_to_client", default=",".join(SSH_ALGO_HMACS)) \
        + ssh_name_list("compression_algorithms_client_to_server", default=",".join(SSH_ALGO_COMPRESSION)) \
        + ssh_name_list("compression_algorithms_server_to_client", default=",".join(SSH_ALGO_COMPRESSION)) \
        + ssh_name_list("languages_client_to_server") \
        + ssh_name_list("languages_server_to_client") \
        + [
        ByteEnumField("kex_first_packet_follows", 0x00, SSH_TYPE_BOOL),
        IntField("reserved", 0x00),
    ]


SSH_DISCONNECT_REASONS = {1: 'HOST_NOT_ALLOWED_TO_CONNECT',
                          2: 'PROTOCOL_ERROR',
                          3: 'KEY_EXCHANGE_FAILED',
                          4: 'RESERVED',
                          5: 'MAC_ERROR',
                          6: 'COMPRESSION_ERROR',
                          7: 'SERVICE_NOT_AVAILABLE',
                          8: 'PROTOCOL_VERSION_NOT_SUPPORTED',
                          9: 'HOST_KEY_NOT_VERIFIABLE',
                          10: 'CONNECTION_LOST',
                          11: 'BY_APPLICATION',
                          12: 'TOO_MANY_CONNECTIONS',
                          13: 'AUTH_CANCELLED_BY_USER',
                          14: 'NO_MORE_AUTH_METHODS_AVAILABLE',
                          15: 'ILLEGAL_USER_NAME',
                          }


class SSHDisconnect(Packet):
    name = "SSH Disconnect"
    fields_desc = [
        IntEnumField("reason", 0xff, SSH_DISCONNECT_REASONS),
        StrCustomTerminatorField(
            "description", "", terminator="\x00\x00\x00\x00"),
        StrCustomTerminatorField(
            "language", "", terminator="\x00", consume_terminator=False),
    ]


class SSH(Packet):
    name = "SSH"

    def is_ascii(s):
        return all(ord(c) < 128 for c in s)

    def guess_payload_class(self, payload):

        try:
            if payload.startswith("SSH-"):
                return SSHIdent

            dummy = SSHMessage(payload, _internal=1)
            if len(payload) <= dummy.length+4:
                return SSHMessage

        except:
            pass
        return SSHEncryptedPacket


# bind magic
bind_layers(TCP, SSH, dport=22)
bind_layers(TCP, SSH, sport=22)

bind_layers(SSH, SSHMessage)
bind_layers(SSHMessage, SSHKexInit, {'type': 0x14})
bind_layers(SSHMessage, SSHDisconnect, {'type': 0x01})
bind_layers(SSH, SSHEncryptedPacket)



try:
    import click
except ImportError:
    raise ImportError(
        "Please install Python dependencies: " "click, colorama (optional)."
    )

from heifip import CONTEXT_SETTINGS, __version__
from heifip.images import NetworkTrafficImage
from heifip.images.flow import FlowImage
from heifip.images.flow_tiled_auto import FlowImageTiledAuto
from heifip.images.flow_tiled_fixed import FlowImageTiledFixed
from heifip.images.markovchain import (MarkovTransitionMatrixFlow,
                                       MarkovTransitionMatrixPacket)
from heifip.images.packet import Packet
from heifip.layers import PacketProcessorType
from heifip.main import Runner


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


@click.version_option(version=__version__)
@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    click.secho("Starting FlowImageProcessor CLI")


_extract_options = [
    click.option(
        "-w",
        "--write",
        "output_dir",
        type=click.Path(),
        required=True,
        help="Destination file path, stores result.",
    ),
    click.option("-r", "--read", "input_dir", required=True, type=click.Path()),
    click.option(
        "-t",
        "--threads",
        "num_threads",
        type=int,
        default=4,
        help="Number of parallel threads that can be used.",
    ),
    click.option(
        "--preprocess",
        "preprocessing_type",
        default="NONE",
        type=click.Choice(
            list(map(lambda x: x.name, PacketProcessorType)), case_sensitive=False
        ),
        help="Applies a preprocessing to the input data:\n none: No preprocessing\n payload: Only payload data is used\n header: Preprocesses headers (DNS,HTTP,IP,IPv6,TCP,UDP supported) to remove some biasing data.",
    ),
    click.option(
        "--min-im-dim",
        "min_image_dim",
        type=int,
        default=0,
        help="Minimum dim ouput images need to have, 0=No minimum dim.",
    ),
    click.option(
        "--max-im-dim",
        "max_image_dim",
        type=int,
        default=0,
        help="Maximum dim ouput images can have, 0=No maximum dim.",
    ),
    click.option(
        "--remove-duplicates",
        "remove_duplicates",
        is_flag=True,
        default=False,
        help="Within a single output folder belonging to a single input folder no duplicate images will be produced if two inputs lead to the same image.",
    ),
]

_flow_options = [
    click.option(
        "--min-packets",
        "min_packets_per_flow",
        type=int,
        default=0,
        help="Minimum packets that a FlowImage needs to have, 0=No minimum packets per flow.",
    ),
    click.option(
        "--max-packets",
        "max_packets_per_flow",
        type=int,
        default=0,
        help="Minimum packets that a FlowImage needs to have, 0=No minimum packets per flow.",
    ),
]

_image_options = [
    click.option(
        "--dim",
        "dim",
        type=int,
        default=8,
        help="Dimension of the image.",
    ),
    click.option(
        "--fill",
        "fill",
        type=int,
        default=0,
        help="Fills remaining parts of the array of the image. Important: value has to be between 0-255.",
    ),
]

_auto_dim_options = [
    click.option(
        "--auto-dim",
        "auto_dim",
        is_flag=True,
        default=False,
        help="Automatically adjust size of image based on the length of the packet/s.",
    ),
]


@cli.group(name="extract", context_settings={"show_default": True})
def extract():
    click.secho("Extract FlowImageProcessor CLI")


@extract.command(name="packet")
@add_options(_extract_options)
@add_options(_flow_options)
@add_options(_image_options)
@add_options(_auto_dim_options)
def extract_packet_image(
    input_dir,
    output_dir,
    num_threads,
    preprocessing_type,
    min_image_dim,
    max_image_dim,
    min_packets_per_flow,
    max_packets_per_flow,
    remove_duplicates,
    dim,
    fill,
    auto_dim,
):
    """Extracts each packet from PCAP file and converts it into a single image representation."""
    runner = Runner(num_threads)
    runner.run(
        input_dir,
        output_dir,
        preprocessing_type,
        PacketImage,
        min_image_dim,
        max_image_dim,
        min_packets_per_flow,
        max_packets_per_flow,
        remove_duplicates,
        dim,
        fill,
        auto_dim,
    )


@extract.command(name="flow")
@add_options(_extract_options)
@add_options(_flow_options)
@add_options(_image_options)
@click.option(
    "--append",
    "append",
    is_flag=True,
    default=False,
    help="",
)
def extract_flow_image(
    input_dir,
    output_dir,
    num_threads,
    preprocessing_type,
    min_image_dim,
    max_image_dim,
    min_packets_per_flow,
    max_packets_per_flow,
    remove_duplicates,
    dim,
    fill,
    append,
):
    """Extracts a list of packets from PCAP file and converts it into an image. You can either append each packet or write each packet into a new line."""
    runner = Runner(num_threads)
    runner.run(
        input_dir,
        output_dir,
        preprocessing_type,
        FlowImage,
        min_image_dim,
        max_image_dim,
        min_packets_per_flow,
        max_packets_per_flow,
        remove_duplicates,
        dim,
        fill,
        append,
    )


@extract.command(name="flow-tiled-fixed")
@add_options(_extract_options)
@add_options(_flow_options)
@add_options(_image_options)
@click.option(
    "--cols",
    "cols",
    type=int,
    default=4,
    help="Number of columns for quadratic representation.",
)
def extract_flow_tiled_fixed_image(
    input_dir,
    output_dir,
    num_threads,
    preprocessing_type,
    min_image_dim,
    max_image_dim,
    min_packets_per_flow,
    max_packets_per_flow,
    remove_duplicates,
    dim,
    fill,
    cols,
):
    """Extracts packets from PCAP file and converts all packets into a single quadratic image based on the number of columns. If more packets are given than the total size of cols*cols, only the first n given packets are used."""
    runner = Runner(num_threads)
    runner.run(
        input_dir,
        output_dir,
        preprocessing_type,
        FlowImageTiledFixed,
        min_image_dim,
        max_image_dim,
        min_packets_per_flow,
        max_packets_per_flow,
        remove_duplicates,
        dim,
        fill,
        cols,
    )


@extract.command(name="flow-tiled-auto")
@add_options(_extract_options)
@add_options(_flow_options)
@add_options(_image_options)
@add_options(_auto_dim_options)
def extract_flow_tiled_fixed_image(
    input_dir,
    output_dir,
    num_threads,
    preprocessing_type,
    min_image_dim,
    max_image_dim,
    min_packets_per_flow,
    max_packets_per_flow,
    remove_duplicates,
    dim,
    fill,
    cols,
):
    """Extracts packets from PCAP file and converts all packets into a single quadratic image. It adjust the size based on the total amount of packets."""
    runner = Runner(num_threads)
    runner.run(
        input_dir,
        output_dir,
        preprocessing_type,
        FlowImageTiledAuto,
        min_image_dim,
        max_image_dim,
        min_packets_per_flow,
        max_packets_per_flow,
        remove_duplicates,
        dim,
        fill,
        cols,
    )


@extract.command(name="markov-flow")
@add_options(_extract_options)
@add_options(_flow_options)
@click.option(
    "--cols",
    "cols",
    type=int,
    default=4,
    help="Number of columns for quadratic representation.",
)
def extract_markov_image(
    input_dir,
    output_dir,
    num_threads,
    preprocessing_type,
    min_image_dim,
    max_image_dim,
    min_packets_per_flow,
    max_packets_per_flow,
    remove_duplicates,
    cols
):
    """Extracts packets from PCAP file and converts it into a quadractic Markov Transition Matrix."""
    runner = Runner(num_threads)
    runner.run(
        input_dir,
        output_dir,
        preprocessing_type,
        MarkovTransitionMatrixFlow,
        min_image_dim,
        max_image_dim,
        min_packets_per_flow,
        max_packets_per_flow,
        remove_duplicates,
        cols
    )

@extract.command(name="markov-packet")
@add_options(_extract_options)
@add_options(_flow_options)
def extract_markov_image(
    input_dir,
    output_dir,
    num_threads,
    preprocessing_type,
    min_image_dim,
    max_image_dim,
    min_packets_per_flow,
    max_packets_per_flow,
    remove_duplicates,
):
    """Extracts packets from PCAP file and converts it into a quadractic Markov Transition Matrix."""
    runner = Runner(num_threads)
    runner.run(
        input_dir,
        output_dir,
        preprocessing_type,
        MarkovTransitionMatrixFlow,
        min_image_dim,
        max_image_dim,
        min_packets_per_flow,
        max_packets_per_flow,
        remove_duplicates,
    )


if __name__ == "__main__":
    cli()



import asyncio
import fnmatch
import glob
import logging
import os
from os.path import dirname, realpath
from queue import Queue
from threading import Thread

from PIL import Image as PILImage
from tqdm import tqdm

from heifip.extractor import FIPExtractor
from heifip.images.flow import FlowImage


class Runner:
    def __init__(self, thread_number) -> None:
        self.thread_number = thread_number
        self.extractor = FIPExtractor()

    def create_image(
        self,
        input_file: str,
        output_dir: str,
        pbar,
        *args
    ):
        imgs = self.extractor.create_image_from_file(input_file, *args)
        pbar.update(1)
        for img in imgs:
            self.extractor.save_image(img, output_dir)

    def start_process(
        self,
        file_queue,
        pbar,
        *args,
    ):
        while not file_queue.empty():
            input_file, output_dir = file_queue.get()
            self.create_image(
                input_file,
                output_dir,
                pbar,
                *args,
            )
            file_queue.task_done()

    def run(
        self,
        input_dir: str,
        output_dir: str,
        *args
    ):

        # Get all executable files in input directory and add them into queue
        file_queue = Queue()
        total_files = 0
        for root, dirnames, filenames in os.walk(input_dir):
            for filename in fnmatch.filter(filenames, "*.pcap"):
                match = os.path.join(root, filename)
                sub_dir = match.replace(input_dir, "")
                file_queue.put((match, f"{output_dir}/{sub_dir}"))
                total_files += 1

        # Start thread
        pbar = tqdm(total=total_files)
        for _ in range(self.thread_number):
            thread = Thread(
                target=self.start_process,
                args=(
                    file_queue,
                    pbar,
                    *args
                ),
            )
            thread.daemon = True
            thread.start()
        file_queue.join()
        pbar.close()



from typing import Type

from scapy.all import Packet, RandMAC
from scapy.layers.inet import Ether


class FIPPacket:
    def __init__(self, packet, address_mapping={}, layer_map={}):
        self.address_mapping = address_mapping
        self.packet = packet

        if layer_map == {}:
            self.layer_map = self.__get_layers()

    def __get_layers(self):
        layer_map = dict()
        layers = self.packet.layers()
        for layer_class in layers:
            layer_map[layer_class] = 1
        return layer_map

    def convert(self, packet_type, packet):
        return packet_type(packet.packet, packet.address_mapping, packet.layer_map)

    def header_preprocessing(self):
        pass

class UnknownPacket(FIPPacket):
    def __init__(self, packet, address_mapping={}, layer_map={}):
        FIPPacket.__init__(self, packet, address_mapping, layer_map)

    def header_preprocessing(self):
        super().header_preprocessing()

class EtherPacket(FIPPacket):
    def __init__(self, packet, address_mapping={}, layer_map={}):
        FIPPacket.__init__(self, packet, layer_map, address_mapping)
        
        if Ether in self.layer_map:
            self.__filter()

    def __filter(self):
        previous_src = self.packet[Ether].src
        previous_dst = self.packet[Ether].dst

        if previous_src in self.address_mapping:
            new_src = self.address_mapping[previous_src]
        else:
            new_src = RandMAC()._fix()
            self.address_mapping[previous_src] = new_src

        if previous_dst in self.address_mapping:
            new_dst = self.address_mapping[previous_dst]
        else:
            new_dst = RandMAC()._fix()
            self.address_mapping[previous_dst] = new_dst

        self.packet[Ether].src = new_src
        self.packet[Ether].dst = new_dst

    def header_preprocessing(self):
        super().header_preprocessing()


import os
from abc import ABC
from enum import Enum, unique
from typing import Type

from scapy.all import (Packet, RandIP, RandIP6, RandMAC, Raw, rdpcap, sniff,
                       wrpcap)
from scapy.layers.dns import DNS
from scapy.layers.http import HTTP, HTTPRequest, HTTPResponse
from scapy.layers.inet import IP, TCP, UDP, Ether
from scapy.layers.inet6 import IPv6

from heifip.exceptions import FIPWrongParameterException
from heifip.layers.dns import DNSPacket
from heifip.layers.http import (HTTPPacket, HTTPRequestPacket,
                                HTTPResponsePacket)
from heifip.layers.ip import IPPacket
from heifip.layers.packet import EtherPacket, FIPPacket, UnknownPacket
from heifip.layers.transport import TransportPacket

__author__ = "Stefan Machmeier"
__copyright__ = "Copyright 2023, heiFIP"
__credits__ = ["Manuel Trageser"]
__license__ = "EUPL"
__version__ = "1.1.0"
__maintainer__ = "Stefan Machmeier"
__email__ = "stefan.machmeier@uni-heidelberg.de"
__status__ = "Production"

SUPPORTED_HEADERS = [IP, IPv6, DNS, HTTPRequest, HTTPResponse, TCP, UDP]


@unique
class PacketProcessorType(Enum):
    NONE = 1
    HEADER = 2


class PacketProcessor:
    def __init__(
        self,
        file_extension="pcap",
    ) -> None:
        pass

    def write_packet(self) -> None:
        # Write pcap
        wrpcap(f"{self.filename}_converted.pcap", self.packets, append=True)

    def read_packets_file(self, file: str, preprocessing_type: PacketProcessorType) -> [FIPPacket]:
        assert os.path.isfile(file)

        # Read PCAP file with Scapy
        packets = []
        # TODO Only read max number of packets
        pcap = sniff(offline=file)
        for pkt in pcap:
            # Start preprocessing for each packet
            processed_packet = self.__preprocessing(pkt, preprocessing_type)
            # TODO Run extract here to reduce amount of loops in code. Atm very inefficient for computation time and memory
            # In case packet returns None
            if processed_packet != None:
                packets.append(processed_packet)
        return packets

    def read_packets_packet(self, packet: [Packet], preprocessing_type: PacketProcessorType) -> [FIPPacket]:
        # Read PCAP file with Scapy
        packets = []
        for pkt in packet:
            # Start preprocessing for each packet
            processed_packet = self.__preprocessing(pkt, preprocessing_type)
            # In case packet returns None
            if processed_packet != None:
                packets.append(processed_packet)
        return packets

    def __preprocessing(self, packet: Packet, preprocessing_type: PacketProcessorType) -> FIPPacket:
        fippacket = UnknownPacket(packet)
        if HTTP in fippacket.layer_map:
            if HTTPRequest in fippacket.layer_map:
                fippacket = fippacket.convert(HTTPRequestPacket, fippacket)
            elif HTTPResponse in fippacket.layer_map:
                fippacket = fippacket.convert(HTTPResponsePacket, fippacket)
            else:
                fippacket = fippacket.convert(HTTPPacket, fippacket)
        elif DNS in fippacket.layer_map:
            fippacket = fippacket.convert(DNSPacket, fippacket)
        elif TCP in fippacket.layer_map or UDP in fippacket.layer_map:
            fippacket = fippacket.convert(TransportPacket, fippacket)
        elif IP in fippacket.layer_map or IPv6 in fippacket.layer_map:
            fippacket = fippacket.convert(IPPacket, fippacket)
        elif Ether in fippacket.layer_map:
            fippacket = fippacket.convert(EtherPacket, fippacket)

        if preprocessing_type == PacketProcessorType.HEADER:
            fippacket.header_preprocessing()

        return fippacket



CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"], show_default=True)

__author__ = "Stefan Machmeier"
__copyright__ = "Copyright 2023, heiFIP"
__credits__ = ["Manuel Trageser"]
__license__ = "EUPL"
__version__ = "1.1.0"
__maintainer__ = "Stefan Machmeier"
__email__ = "stefan.machmeier@uni-heidelberg.de"
__status__ = "Production"